import re, string
import sre_parse

MAXREPEAT = sre_parse.MAXREPEAT

class Node:
    def __add__(self, other):
        return Seq( (self, other) )
    def __or__(self, other):
        return Alt( (self, other) )
    def group_names(self):
        return ()
    def _select_names(self, names):
        # Don't call this function - subtrees can be shared,
        # so you need to copy first before selecting
        pass
    def copy(self):
        # This is actually a deep copy
        raise NotImplementedError
    def __str__(self):
        # Shows the re string
        raise NotImplementedError
    def make_parser(self):
        import Generate
        return Generate.generate(self)

class NodeList(Node):
    def group_names(self):
        names = {}
        for exp in self.expressions:
            for name in exp.group_names():
                names[name] = 1
        return tuple(names.keys())
    def _select_names(self, names):
        for exp in self.expressions:
            exp._select_names(names)
    def copy(self):
        return self.__class__(map(lambda x: x.copy(), self.expressions))

# A set of expressions: 'a|b|c'
class Alt(NodeList):
    def __init__(self, expressions):
        self.expressions = expressions
    def __or__(self, other):
        if isinstance(other, Alt):
            return Alt(self.expressions + other.expressions) # use tuples!
        else:
            return Alt(self.expressions + (other,))
    def __str__(self):
        return string.join(map(str, self.expressions), '|')

# A sequence of expressions: '[ab][cd][ef]'
class Seq(NodeList):
    def __init__(self, expressions):
        self.expressions = expressions
    def __add__(self, other):
        if isinstance(other, Seq):
            return Seq(self.expressions + other.expressions) # use tuples!
        return Seq(self.expressions + (other,))
    def __str__(self):
        return string.join(map(str, self.expressions), "")
    

# taken from re.escape, except also don't escape " "
def escape(pattern):
    "Escape all non-alphanumeric characters in pattern."
    result = list(pattern)
    alphanum=string.letters+'_'+string.digits+" "
    for i in range(len(pattern)):
        char = pattern[i]
        if char not in alphanum:
            if char=='\000': result[i] = '\\000'
            else: result[i] = '\\'+char
    return string.join(result, '')

# A single character: 'a'
class Literal(Node):
    def __init__(self, char, not_flag = 0):
        self.char = char
        self.not_flag = not_flag
    def copy(self):
        return Literal(self.char, self.not_flag)
    def __str__(self):
        c = escape(self.char)
        if self.not_flag:
            return '[^%s]' % c
        return c

# A sequence of characters: 'abcdef'
class Str(Node):
    def __init__(self, string):
        self.string = string
    def copy(self):
        return Str(self.string)
    def __str__(self):
        return escape(self.string)

# A range of characters: '[abc]'
class In(Node):
    def __init__(self, chars, not_flag):
        self.chars = chars
        self.not_flag = not_flag
    def copy(self):
        return In(self.chars, self.not_flag)
    def __str__(self):
        if self.not_flag:
            return '[^%s]' % _minimize_range(self.chars)
        else:
            return '[%s]' % _minimize_range(self.chars)

# A grouping: '(?P<name>expression)'
#             '(?:expression)'
#             '(expression)'    -- same as (?:...) since I don't track
#                                  or use the group number


# Group names must be valid XML identifiers
def _verify_name(s):
    assert s, "Group name can not be the empty string"
    if not sre_parse.isname(s):
        raise AssertionError, "Illegal character in group name %s" % repr(s)
    return s

class Subpattern(Node):
    def __init__(self, name, expression):
        self.name = name
        self.expression = expression
    def group_names(self):
        if self.name is not None:
            return (self.name, ) + self.expression.group_names()
        return self.expression.group_names()
    def _select_names(self, names):
        if self.name is not None and self.name not in names:
            self.name = None
        self.expression._select_names(names)
    def copy(self):
        return Subpattern(self.name, self.expression.copy())
    def __str__(self):
        if self.name is None:
            return '(%s)' % str(self.expression)
        else:
            return '(?P<%s>%s)' % (_verify_name(self.name),
                                   str(self.expression))
        
# Greedy repeat: 'a*'
#                'a{3,5}'
#                'a+'
class MaxRepeat(Node):
    def __init__(self, expression, min_count=0,
                 max_count=MAXREPEAT):
        self.expression = expression
        self.min_count = min_count
        self.max_count = max_count
    def group_names(self):
        return self.expression.group_names()
    def copy(self):
        return MaxRepeat(self.expression.copy(), self.min_count,
                         self.max_count)
    def _select_names(self, names):
        self.expression._select_names(names)
    def __str__(self):
        min_count = self.min_count
        max_count = self.max_count
        subexp = self.expression
        if isinstance(subexp, NodeList):
            need_group = 1
        elif isinstance(subexp, Str):
            need_group = 1
        else:
            need_group = 0

        if need_group:
            s = "(%s)" % str(subexp)
        else:
            s = str(subexp)
        if type(min_count) == type("") or type(max_count) == type(""):
            if min_count == max_count:
                ext = "{%s}" % min_count
            else:
                ext = "{%s,%s}" % (min_count, max_count)
        else:
            if min_count == 0 and max_count == MAXREPEAT:
                ext = "*"
            elif min_count == 1 and max_count == MAXREPEAT:
                ext = "+"
            elif min_count == 0 and max_count == 1:
                ext = "?"
            elif min_count == max_count:
                assert 0 <= min_count, \
                       "repeat count must be non-negative, not %s" % min_count
                ext = "{%d}" % max_count
            else:
                assert 0 <= min_count < max_count, \
                       "illegal repeat count (%d, %d)" % (min_count, max_count)
                ext = "{%d,%d}" % (min_count, max_count)

        return s + ext

# At the beginning of the string: '^'
class AtBeginning(Node):
    def copy(self):
        return AtBeginning()
    def __str__(self):
        return '^'

# At the end of the string: '$'
class AtEnd(Node):
    def copy(self):
        return AtEnd()
    def __str__(self):
        return '$'
    
# Any character except newline: '.'
class Any(Node):
    def copy(self):
        return Any()
    def __str__(self):
        return '.'

# Lookahead assertions: '(?=...)'
#                       '(?!...)'
class Assert(Node):
    def __init__(self, expression, not_flag = 0):
        self.expression = expression
        self.not_flag = not_flag
    def copy(self):
        return Assert(self.expression.copy(), self.not_flag)
    def __str__(self):
        if self.not_flag:
            return '(?!%s)' % str(self.expression)
        else:
            return '(?=%s)' % str(self.expression)

# group reference: '(.)\1'
class GroupRef(Node):
    def __init__(self, name):
        assert type(name) == type('')
        self.name = name
    def copy(self):
        return GroupRef(self.name)
    def __str__(self):
        return "(?P=%s)" % self.name
        
_minimize_escape_chars = {
    "\a": r"\a",
    "\b": r"\b",
    "\n": r"\n",
    "\r": r"\r",
    "\t": r"\t",
    "\f": r"\f",
    "\v": r"\v",
    " ": " ",
    }
    
def _minimize_escape_char(c):
    x = _minimize_escape_chars.get(c)
    if x is None:
        return escape(c)
    return x

def _minimize_escape_range(c1, c2):
    if c1 == '0' and c2 == '9':
        return r"\d"
    return _minimize_escape_char(c1) + '-' + _minimize_escape_char(c2)
    

def _minimize_range(s):
    if not s:
        return s
    has_hyphen = 0
    if '-' in s:
        has_hyphen = 1
        s = string.replace(s, "-", "")

    chars = list(s)
    chars.sort()
    unique = []
    prev_c = None
    for c in chars:
        if c != prev_c:
            unique.append(c)
            prev_c = c
    
    s = string.join(unique, "")
    t = ""
    prev = None
    prev_pos = 0
    pos = 0
    for c in unique:
        val = ord(c)
        
        if val - 1 != prev:
            # either beginning of string or non-sequential
            if prev is None:
                # beginning of string
                prev_pos = 0
            else:
                # non-sequential
                if prev_pos == pos - 1:
                    t = t + _minimize_escape_char(s[prev_pos])
                elif prev_pos == pos - 2:
                    t = t + _minimize_escape_char(s[prev_pos]) + \
                            _minimize_escape_char(s[prev_pos+1])
                else:
                    t = t + _minimize_escape_range(s[prev_pos], s[pos-1])
                prev_pos = pos
        else:
            # sequential, so just advance
            pass

        prev = val
        pos = pos + 1

    if s:
        if prev_pos == pos - 1:
            t = t + _minimize_escape_char(s[prev_pos])
        elif prev_pos == pos - 2:
            t = t + _minimize_escape_char(s[prev_pos]) + \
                    _minimize_escape_char(s[prev_pos+1])
        else:
            t = t + _minimize_escape_range(s[prev_pos], s[pos-1])
    else:
        # was only the hyphen
        pass
    if has_hyphen:
        t = t + '-'
    return t


def _test_minimize():
    data = (
        ("a", "a"),
        ("ab", "ab"),
        ("abc", "a-c"),
        ("abcdef", "a-f"),
        ("abcj", "a-cj"),
        ("abcjk", "a-cjk"),
        ("abcjkl", "a-cj-l"),
        ("0123456789", r"\d"),
        (string.lowercase, r"a-z"),
        (string.lowercase + string.uppercase, r"A-Za-z"),
        (string.lowercase + string.uppercase + string.digits, r"\dA-Za-z"),
        ("\007\010", r"\a\b"),
        ("1357", "1357"),
        ("", ""),
        ("\\", "\\\\"),
        ("-0123456789", "\\d-"),
        ("-", "-"),
        ("^", "\\^"),
        ("^-", "\\^-"),
        ("[]", "\\[\\]"),
        ("\001", "\\\001"),
        ("\001\002\003\004\005\006\007", "\\\001-\\a"),
        )
    for input, output in data:
        x = _minimize_range(input)
        if x != output:
            print repr(input), ":", repr(output), "!=", repr(x)

def _test_group_names():
    a = Literal('a')
    b = Literal('b')
    g1 = Subpattern("foo", a)
    g2 = Subpattern("bar", b)
    g = Subpattern("spam", Alt([g1, g2]))
    assert str(g) == "(?P<spam>(?P<foo>a)|(?P<bar>b))", str(g)
    assert g.group_names() == ("spam", "foo", "bar"), g.group_names()

    h = g.copy()
    assert str(h) == "(?P<spam>(?P<foo>a)|(?P<bar>b))", str(h)
    
    g2.name = "baz"
    assert g.group_names() == ("spam", "foo", "baz"), g.group_names()

    assert str(h) == "(?P<spam>(?P<foo>a)|(?P<bar>b))", str(h)

    g._select_names(["foo", "baz"])
    assert g.group_names() == ("foo", "baz"), g.group_names()
        

def _test():
    _test_minimize()
    _test_group_names()
