# Create the tables needed for mxTextTools

import string
import TextTools
TT = TextTools
import sre_parse
import Expression, convert_re

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# HACK TO SUPPORT back references using mxTextTools!

# THIS MEANS SINGLE THREADED SUPPORT for anything using
# backreferences!  There is a much more complicated solution where the
# first element of any taglist is defined to contain the _match_group
# for that parse session.  I don't want to do that, since another
# approach is to modify mxTextTools to pass around an extra state
# object, or to write my own code.  (Did someone say NIMBY? :)
_match_group = {}


def _do_callback(s, begin, end, taglist, callback):
    for item in taglist:
        tag, l, r, subtags = item
        if begin < l:
            callback.characters(s, begin, l-begin)
        else:
            assert begin == l, "begin = %d and l = %d" % (begin, l)

        if tag != ">ignore":
            callback.startElement(tag, {})

        if subtags:
            _do_callback(s, l, r, subtags, callback)
        else:
            callback.characters(s, l, r-l)
        begin = r

        if tag != ">ignore":
            callback.endElement(tag)

    if begin < end:
        callback.characters(s, begin, end-begin)

import xml.sax.saxlib
class StateTableException(xml.sax.saxlib.SAXException):
    pass
class StateTableEOFException(StateTableException):
    pass

# This needs an interface like the standard XML parser
class StateTable:
    def __init__(self, taglist = None):
        if taglist is None:
            self.taglist = []
        else:
            self.taglist = taglist

        self._cont_handler = None
        self._err_handler = None

        self._want_groupref_names = []
        
    def getContentHandler(self):
        return self._cont_handler
    def setContentHandler(self, handler):
        self._cont_handler = handler

    def getErrorHandler(self):
        return self._err_handler
    def setErrorHandler(self, handler):
        self._err_handler = handler
        
    def __str__(self):
        x = StringIO()
        pprint.pprint(self.taglist, x)
        return x.getvalue()

    def parse(self, s):
        if self._want_groupref_names:
            _match_group.clear()
        result, taglist, pos = TT.tag(s, tuple(self.taglist), 0, len(s))

        cont = self._cont_handler
        err = self._err_handler
        if cont is not None:
            cont.startDocument()
            # There should be a start Element with the Martel interface
            # version number.
            _do_callback(s, 0, pos, taglist, cont)
        if not result:
            if err is not None:
                err.fatalError(StateTableException(
                    "error parsing beyond character %d" % pos,
                    None
                    ))
        elif pos != len(s):
            if err is not None:
                err.error(StateTableEOFException(
                    "parsing finished too early, at character %d, not %d" % \
                    (pos, len(s)),
                    None
                    ))
        else:
            if cont is not None:
                cont.endDocument()

# the generate_* functions all create a StateTable which expects to
# start at the first element and succeed by falling one past the last
# element.

def generate_branch(expression, groupref_names):
    tables = []
    for expr in expression.expressions:
        tables.append(_generate(expr, groupref_names))
    i = 0
    n = len(tables)
    combined = StateTable()
    for table in tables:
        combined.taglist.append( \
            (None, TT.SubTable, tuple(table.taglist), +1, n-i+1)
            )
        i = i + 1
    combined.taglist.append( (None, TT.Fail, TT.Here) )
    return combined

def generate_seq(expression, groupref_names):
    combined = StateTable()
    for exp in expression.expressions:
        table = _generate(exp, groupref_names)
        combined.taglist.extend(table.taglist)
    return combined
    
def generate_literal(expression, groupref_names):
    if expression.not_flag:
        # Can't use "InNot" since it can match EOF
        return StateTable( [(None, TT.IsIn,
                             convert_re.invert(expression.char))] )
    else:
        return StateTable( [(None, TT.Is, expression.char)] )

def generate_string(expression, groupref_names):
    return StateTable( [(None, TT.Word, expression.string)] )

def generate_in(expression, groupref_names):
    if expression.not_flag:
        # I don't use "IsNotIn" since that allows EOF
        return StateTable( [(None, TT.IsIn,
                             convert_re.invert(expression.chars))] )
    else:
        return StateTable( [(None, TT.IsIn, expression.chars)] )

class SetSubpatternValue:
    def __init__(self, name):
        self.name = name
    def __call__(self, taglist, text, l, r, subtags):
        taglist.append( (self.name, l, r, subtags) )
        #print self.name, "sees", repr(text[l:r]), l, r
        _match_group[self.name] = text[l:r]

def generate_subpattern(expression, groupref_names):
    name = expression.name
    if name is None:
        # mxTextTools won't let me create subtags if there is no
        # parent tag, so I create a tag which is illegal, only
        # to ignore later.
        name = '>ignore'
        
    table = _generate(expression.expression, groupref_names)
    if groupref_names.get(name) != 1:
        return StateTable( [(name, TT.Table, tuple(table.taglist)) ] )
    else:
        return StateTable( [(SetSubpatternValue(name), TT.Table+TT.CallTag,
                             tuple(table.taglist)), ] )

class HandleRepeatCount:
    def __init__(self, tagtable, min_count, max_count):
        self.tagtable = tagtable
        self.min_count = min_count
        self.max_count = max_count

        self.taglist = None

    def _get_ranges(self):
        min_count = self.min_count
        if type(min_count) == type(""):
            min_count = _match_group[min_count] # group must already exist!
            min_count = string.atoi(min_count)  # requires an integer!
        
        max_count = self.max_count
        if type(max_count) == type(""):
            max_count = _match_group[max_count] # group must already exist!
            max_count = string.atoi(max_count)  # requires an integer!

        return min_count, max_count
    
    def __call__(self, text, x, end):
        # Called by 'Call' to detect a match.
        # I do the full match here and store the results for later use.
        # If successful, I return x+1, else return x+0
        min_count, max_count = self._get_ranges()
        assert min_count == max_count, \
               "cannot have different sizes: %s %s" % (min_count, max_count)
        tagtable = self.tagtable * min_count
        result, taglist, pos = TT.tag(text, tagtable, x, end)
        if result == 1:
            #print "\nSuccess on", self.min_count
            #print repr(text[x:pos])
            self.taglist = taglist
            return pos + 1  # +1 because {0} is allowed; Skip -1 later
        else:
            self.taglist = None
            return x
        
    def calltag(self, taglist, text, l, r, subtags):
        # Called by 'CallTag' to say I matched the above __call__
        assert not subtags, repr(subtags)
        # The '-1' is because of the hack around having a successful
        # zero size match
        taglist.append(">ignore", l, r-1, self.taglist)
        
    
def generate_named_max_repeat(expression, groupref_names):
    if type(expression.min_count) != type("") or \
       type(expression.max_count) != type(""):
        raise NotImplementedError("Cannot mix numeric and named repeat counts")
    if expression.min_count != expression.max_count:
        raise NotImplementedError("Only a single named repeat count allowed")
    
    table = _generate(expression.expression, groupref_names)
    counter = HandleRepeatCount(tuple(table.taglist),
                                expression.min_count,
                                expression.max_count)
    return StateTable(
        [(counter.calltag, TT.Call + TT.CallTag, counter, TT.MatchFail), 
         (None, TT.Skip, -1, TT.MatchFail),
         ])

def generate_max_repeat(expression, groupref_names):
    expr = expression.expression
    min_count = expression.min_count
    max_count = expression.max_count

    if type(min_count) == type("") or type(max_count) == "":
        return generate_named_max_repeat(expression, groupref_names)

    assert 0 <= min_count <= max_count, "bad ranges (%sd, %d)" % \
           (min_count, max_count)

    table = _generate(expr, groupref_names)
    combined = StateTable()

    for i in range(min_count):
        combined.taglist.append( (None, TT.SubTable, tuple(table.taglist)) )

    if max_count == sre_parse.MAXREPEAT:
        combined.taglist.append( (None, TT.SubTable, tuple(table.taglist),
                                  +1, 0))
    elif min_count == max_count:
        pass
    else:
        offset = max_count - min_count
        for i in range(offset):
            combined.taglist.append( (None, TT.SubTable,
                                      tuple(table.taglist),
                                      +offset, +1) )
            offset = offset - 1
    return combined

# XXX Is this correct?  This is the multiline behaviour
def check_at_beginning(text, x, end):
    if x == 0:
        return x
    if text[x-1] == "\n":
        return x
    return x + 1

def generate_at_beginning(expression, groupref_names):
    return StateTable( [(None, TT.Call, check_at_beginning, +2, +1),
                        (None, TT.Skip, -1, TT.MatchFail, TT.MatchFail),
                        ])
def generate_at_end(expression, groupref_names):
    return StateTable( [(None, TT.EOF, TT.Here)] )

_any_char = string.join(map(chr,
                            range(0, ord("\n")) + range(ord("\n")+1, 256)
                            ), "")
def generate_any(expression, groupref_names):
    return StateTable( [(None, TT.IsIn, _any_char), ] )

# workaround an mxTextTools bug?
class CheckAssertNotWrapper:
    def __init__(self, tag_words):
        self.tag_words = tag_words
    def __call__(self, text, x, end):
        pos = check_assert_not(text, x, end, self.tag_words)
        return pos

def check_assert_not(text, x, end, tag_words):
    result, taglist, pos = TT.tag(text, tag_words, x, end)
    if result:
        # This failed
        return x
    # On success, move forward 1, to be removed later
    return x + 1


# workaround an mxTextTools bug?
class CheckAssertWrapper:
    def __init__(self, tag_words):
        self.tag_words = tag_words
    def __call__(self, text, x, end):
        pos = check_assert(text, x, end, self.tag_words)
        return pos

def check_assert(text, x, end, tag_words):
    result, taglist, pos = TT.tag(text, tag_words, x, end)
    if result:
        # This succeeded, move forward 1, to be removed later
        return x+1
    # failed
    return x

def generate_assert(expression, groupref_names):
    table = _generate(expression.expression, groupref_names)
    if expression.not_flag:
        return StateTable( [
            (None, TT.Call, CheckAssertNotWrapper(tuple(table.taglist)),
             TT.MatchFail),
            (None, TT.Skip, -1, TT.MatchFail),
            ])
    else:
        return StateTable( [
            (None, TT.Call, CheckAssertWrapper(tuple(table.taglist)),
             TT.MatchFail),
            (None, TT.Skip, -1, TT.MatchFail),
            ])

class CheckGroupRef:
    def __init__(self, name):
        self.name = name
    def __call__(self, text, x, end):
        match_text = _match_group[self.name]  # group name not yet found
        #print repr(text[x:x+len(match_text)]), repr(match_text)
        if text[x:x+len(match_text)] != match_text:
            #print "Return", x
            return x
        #print "RETURN", x+len(match_text)+1
        return x+len(match_text)+1

def generate_groupref(expression, groupref_names):
    # Look up the string from the match group
    # It can be of length 0 or more, so the call returns +length+1 for
    # match and +0 for no match.
    return StateTable( [
        (expression.name, TT.Call, CheckGroupRef(expression.name),
         TT.MatchFail),
        (None, TT.Skip, -1, TT.MatchFail),
        ] )
    
generate_table = {
    Expression.Alt: generate_branch,
    Expression.Seq: generate_seq,
    Expression.Literal: generate_literal,
    Expression.Str: generate_string,
    Expression.In: generate_in,
    Expression.Subpattern: generate_subpattern,
    Expression.MaxRepeat: generate_max_repeat,
    Expression.AtBeginning: generate_at_beginning,
    Expression.AtEnd: generate_at_end,
    Expression.Any: generate_any,
    Expression.Assert: generate_assert,
    Expression.GroupRef: generate_groupref,
}
    

# The internal recursive call
def _generate(expression, groupref_names):
    func = generate_table[expression.__class__]
    return func(expression, groupref_names)

def generate(expression):
    groupref_names = _find_groupref_names(expression)
    obj = _generate(expression, groupref_names)
    obj._want_groupref_names = groupref_names
    return obj

def _find_groupref_names(expression):
    want_names = {}
    if isinstance(expression, Expression.Alt) or \
       isinstance(expression, Expression.Seq):
        for x in expression.expressions:
            want_names.update(_find_groupref_names(x))
    elif isinstance(expression, Expression.Subpattern) or \
         isinstance(expression, Expression.Assert):
        want_names.update(_find_groupref_names(expression.expression))
    elif isinstance(expression, Expression.MaxRepeat):
        if type(expression.min_count) == type(""):
            want_names[expression.min_count] = 1
        if type(expression.max_count) == type(""):
            want_names[expression.max_count] = 1
        want_names.update(_find_groupref_names(expression.expression))
    elif isinstance(expression, Expression.GroupRef):
        want_names[expression.name] = 1
    elif isinstance(expression, Expression.Literal) or \
         isinstance(expression, Expression.Str) or \
         isinstance(expression, Expression.In) or \
         isinstance(expression, Expression.AtBeginning) or \
         isinstance(expression, Expression.AtEnd) or \
         isinstance(expression, Expression.Any):
        pass
    else:
        raise NotImplementedError, "What is a %s?" % repr(expression)
    return want_names

def _test():
    import convert_re
    import re

    from xml.sax import saxlib
    class _Test(saxlib.HandlerBase):
        def __init__(self):
            self.good_parse = 0
        def startDocument(self):
            self.good_parse = 0
        def fatalError(self, exc):
            pass
        def error(self, exc):
            # Called when there aren't enough characters
            self.good_parse = 1
        def endDocument(self):
            self.good_parse = 1
    cb = _Test()
    
    patterns = (
        ("a", ("a",), ("A", "", "Z")),
        ("[a-z]", ("a", "b", "q"), ("A", "-")),
        ("[^abc]", ("A", "d", "fab"), ("a", "b", "c", "ab", "")),
        ("a+", ("a", "aaa"), ("A", "baa")),
        ("a*", ("", "a", "aaa"), ()),
        ("a*$", ("a", "aaa"), ("A", "baa", "aaaaab")),
        ("(ab|ac)", ("ab", "ac"), ("aa", "A", "a", "cb")),
        ("(ab|ac)*$", ("", "ab", "ac", "abacabac", "ababab"),
                      ("aa", "A", "a", "cb", "ababababaca")),
        ("ab{3}$", ("abbb",), ("abb", "bbb", "abbbb")),
        ("ab{3,}$", ("abbb", "abbbb", "abbbbbbbbb"), ("abb", "bbb", "abbbc")),
        ("ab{3,}", ("abbb", "abbbb", "abbbbbbbbb", "abbbc"), ("abb", "bbb")),
        ("abc$|abcd|bc|d", ("abc", "abcd", "bc", "d", "abcdef"),
                           ("xabc", "ab", "a", "", "abce")),
        ("^a", ("a", "aa"), ("b", "ba", "c", "")),
        ("^[^b]", ("a", "aa", "c"), ("b", "ba", "")),
        ("a(?!b).", ("aa", "ac", "aab"), ("a", "ab", "abc")),
        ("a(?=[bc])..", ("abx", "acx", "aba"), ("ac", "ab", "adb")),

        ("ab?", ("a", "ab", "abb", "ac"), ("", "cab", "x")),
        ("ab{2,4}", ("abb", "abbb", "abbbb", "abbbbc", "abbbbbb"),
         ("ab", "abc", "xabbb")),
        ("ab{2,4}$", ("abb", "abbb", "abbbb"),
         ("ab", "abc", "xabbb", "abbbbc", "abbbbbb")),
        ("ab{2,4}c", ("abbc", "abbbc", "abbbbc", "abbbbcd"),
         ("abc", "abbbbbc", "abcbbc")),
        ("ab?c", ("ac", "abc"), ("abb", "abbc", "abbbc")),
        )
    for re_pat, good_list, bad_list in patterns:
        print "Testing", repr(re_pat)
        exp = _generate(convert_re.make_expression(re_pat))
        exp.setContentHandler(cb)
        exp.setErrorHandler(cb)
        pat = re.compile(re_pat)
        for word in good_list:
            exp.parse(word)
            if not cb.good_parse:
                print "  Problem not recognizing", repr(word)
            if not pat.match(word):
                print "  Re problem recognizing", repr(word)
        for word in bad_list:
            exp.parse(word)
            if cb.good_parse:
                print "  Should not recognize", repr(word)
                print "    taglist is", exp.taglist
            if pat.match(word):
                print "  Re should not recognize", repr(word)
